# This code is copyrighted by Stephan Richter in August, 2000.
# This code is copyrighted by iuveno AG in January, 2001.
#
# written by Stephan Richter
# Last updated: February 4, 2001

import Globals, General
from Globals                  import HTMLFile
from Products.ZSQLMethods.SQL import SQL
from OFS.ObjectManager        import ObjectManager
from ZPublisher.HTTPResponse  import HTTPResponse

from   DBObjectRoleSupport    import DBObjectRoleSupport
from   DBObjectCopySupport    import DBObjectCopySupport
from   DBObjectUndoSupport    import DBObjectFolderUndoSupport
from   DBObjectSearchSupport  import DBObjectFolderSearchSupport
from   ActionTriggerSupport   import ActionTriggerSupport
from   RDB2ObjectSupport      import RDB2ObjectSupport
import DBObjectExceptions
from types                    import ListType, TupleType, StringType

from OFS.SimpleItem import Item

import string


class DBObjectFolder(ObjectManager, Item, DBObjectFolderSearchSupport, DBObjectFolderUndoSupport,
                     DBObjectCopySupport, DBObjectRoleSupport, RDB2ObjectSupport, ActionTriggerSupport):
    '''DBObjectFolder class - Abstract description of the DBObject container'''

    isVolatile           = 0
    isSQLMethodContainer = 0
    
    # This dictionary contains meta-type - objects pairs of all the allowed sub-objects
    allowedSubObjects = ()
    genericSubObject = None
    
    # this variable contains the fingerprint of all SQL Methods that need to be initialized
    sqlMethods = ()
    
    # variables that must be defined for Zope
    meta_type = None
    icon = None
    
    # set the permissions required for executing a certain management method
    __ac_permissions__ = ( (
        ('View management screens',     ('manage_main', 'manage_addForm')),
        ('Access contents information', ('objectIds', 'objectValues', 'objectItems', ''), ('Anonymous', 'Manager')),
        ('View DBObjects',              ('getAllSubObjects', 'getAllowedSubObjects', 'objectItems', 'objectValues',
                                         'objectIds', 'tpURL',)),
        ('Add DBObjects',               ('manage_add', 'findExistingSubObjects',)),
        ('Delete DBObjects',            ('manage_delObjects', 'deleteDBAttributes',)), )
        + DBObjectRoleSupport.__ac_permissions__ )


    # The folder will always have a sub-object overview
    manage_options = (({'label':'Contents',    'action':'manage_main'},)
                      + DBObjectRoleSupport.manage_options
                      + DBObjectFolderSearchSupport.manage_options)
    
    # Form providing fields of the attributes required to initialize a sub-object 
    # This method will be overwritten by the derived class
    manage_addForm = HTMLFile('dtml/dbObject_manageAddForm', globals())
    manage_main = HTMLFile('dtml/folder_manageContents', globals())

    manage_pasteObjects = DBObjectCopySupport.manage_pasteObjects


    def __bobo_traverse__(self, REQUEST, name):
        '''__bobo_traverse__(self, REQUEST, name) --> traverse through the object; return object or HTML'''

        # first see, whether the object with the id name exists in here
        try: return self._getObjectSafely(name)
        except 'NotFound': pass

        # if it is not an object, try whether it is an attribute
        try: return getattr(self, name)
        except AttributeError: pass
            
        # the object was not found, so output a standard HTML Not Found Error
        REQUEST.RESPONSE.notFoundError(name)


    persistent__getitem__ = ObjectManager.__getitem__
    def __getitem__(self, item):
        '''__getitem__(self, item) --> try to get the item in a dictionary-like syntax; return object'''
        try:
            if (type(item) is StringType): return self._getObjectSafely(item)
            else: raise TypeError, 'The index must be a string.'
        except:
            return self.persistent__getitem__(item)



    def __init__(self, parent, id, title=''):
        '''__init__(self, parent, id=None, title=None) --> Initialize the object; return None'''

        self._v_checked = 0

	if self.isSQLMethodContainer: 
	    sqlContainer = self
	else: 
	    sqlContainer = self._getSQLMethodContainer(parent)

	if self.isVolatile:
	    self.sqlContainer = sqlContainer

	for subObjectMetaType in self.allowedSubObjects:
	    subObject = General.getSubObjectClass(subObjectMetaType)
	    subObject.dbTable.createBasicSQLMethods(sqlContainer)
	    if self.useUndo:
		self.initUndoSQLMethods(sqlContainer, subObject)

	self.initSQLMethods(sqlContainer)
        
        # Ensure that the id is a string
        self.id = str(id)
        self.title = title
   

    def _getSQLMethodContainer(self, object):
        '''_getSQLMethodContainer(self, metaType, object) --> find the SQL Method Container; return object'''

        if self.isVolatile and hasattr(self, 'sqlContainer'):
            return self.sqlContainer
        elif self.isVolatile:
            return object._getSQLMethodContainer(object)
        else:
            if object.isSQLMethodContainer:
                return object
            else:
                if object.getParentNode() is not None: return self._getSQLMethodContainer(object.aq_parent)
                else: raise 'SQL Method Container not found!!! Yikes!'


    def getSubObjectClass(self, metaType):
        '''getSubObjectClass(self, metaType) --> return the class definition of the metaType; return class'''
        return General.getSubObjectClass(metaType)
    

    def all_meta_types(self):
        '''all_meta_types(self) --> Get all the allowed meta types'''
        product = string.split(str(self.__class__.__module__), '.')[1]
        meta_types = []
        for meta_type in self.allowedSubObjects:
            meta_types.append({'name'       : meta_type,
                               'permission' : 'Add DBObject',
                               'product'    : product,
                               'action'     : 'manage_addForm'})
        return meta_types
            

    def getDBContainerInfo(self):
        '''getDBContainerInfo(self) --> get id & meta type of DB Container; return tuple of strings'''

        if hasattr(self, 'getDBContainerOfSubObjects'):
            container     = self.getDBContainerOfSubObjects()
            containerId   = container.id
            containerType = container.meta_type
        else:
            containerId   = 0
            containerType = ''

        return (containerId, containerType)

    
    def manage_add(self, REQUEST, returnId=0, subObjectMetaType=None, returnObj=0):
        '''manage_add(self, REQUEST, returnId = 0, subObjectMetaType = None) --> add a sub-object
           to the DBObjectFolder; return None or new sub-object id'''

        self._beforeAdd(self, REQUEST)

        dbContainer = self.getDBContainerInfo()
        REQUEST['containerid']=dbContainer[0]
        REQUEST['containertype']=dbContainer[1]
    
        # try to find a meta type, if none was passed
        if not subObjectMetaType:
            if len(self.allowedSubObjects):
                subObjectMetaType = self.allowedSubObjects[0]
            else:
                raise DBObjectException.NoAllowedSubObjectMetaTypeFound()
    
        # check whether the passed sub-object meta type is an allowed one; if not
        # raise an IllegalSubObjectMetaType exception
        if subObjectMetaType in self.allowedSubObjects:
            # Here is the cool stuff: use eval to generalize the add method
            objectClass = General.getSubObjectClass(subObjectMetaType)
            s = objectClass(self, REQUEST)
            
        else:
            raise DBObjectException.IllegalSubObjectMetaType(self.allowedSubObjectMetaTypes,
                                                             subObjectMetaType)
        
        # add the sub-object to this object (folder)
        self._setObject(str(s.id), s) 
        if not s.isVolatile:
            self[s.id]._afterAdd(self, REQUEST)
            # Create all possible subfolders
            if hasattr(s, 'allowedSubObjects') and s.allowedSubObjects:
                for metaType in s.allowedSubObjects:
                    self[s.id].manage_add(subObjectMetaType=metaType)

        # Return the new spec id, if returnID is true
        if returnId: 
            return str(s.id)

        # Support returnObj for orthogonality
        if returnObj:
            return self[s.id]

        # Try to go back to the folder's default (main) view
        if REQUEST is not None:
            return self.manage_main(self, REQUEST)
    
    
    def getAllSubObjects(self, metaType=None, orderBy=''):
        '''getAllSubObjects(self) --> retrieve all the sub-objects from the DB; return db rows'''
        if not metaType:
            metaType = self.genericSubObject
        if not metaType:
            metaType = self.allowedSubObjects[0]

        dbContainer = self._getSQLMethodContainer(self)
        containerId, containerType = self.getDBContainerInfo()

        try:
            return dbContainer.qry_getAllSubObjects(table=metaType, containerid=containerId,
                                                    containertype=containerType, orderBy=orderBy)
        except:
            return []
        

    def getAllowedSubObjects(self, user, metaType=None, orderBy=''):
        '''getAllowedSubObjects(self, metaType=None, orderBy='') --> get only the objects you have permissions for;
           return list of objects'''

        if not metaType:
            metaType = self.allowedSubObjects[0]

        dbRows  = self.getAllSubObjects(metaType, orderBy)
        objects = self.convertSQLResultToObjects(dbRows, metaType)

        allowedObjects = []
        for object in objects:
            if user.canAccessObject(object, type='view'):
                allowedObjects.append(object)

        return allowedObjects
        

    def initSQLMethods(self, container, connectionId='dbConnection'):
        '''initSQLMethods(self, container) --> initializes additional SQML Methods; return None'''

        for sqlMethod in DBObjectFolderSearchSupport.sqlMethods:
            if not hasattr(container, sqlMethod[0]):
                setattr(container, sqlMethod[0], SQL(sqlMethod[0], '', connectionId, sqlMethod[1], sqlMethod[2]))

        # for every specified SQL Method, check whether it is already in the container; if not, generate
        # the SQL Method and add it to the container
        for sqlMethod in self.sqlMethods:
            if not hasattr(container, sqlMethod[0]):
                setattr(container, sqlMethod[0], SQL(sqlMethod[0], '', connectionId, sqlMethod[1], sqlMethod[2]))


    def findExistingSubObjects(self, REQUEST=None):
        '''findExistingSubObjects(self, REQUEST=None) --> Find the sub-objects in the database and
           adds them to Zope; return None'''

        if not self.isVolatile:

            # check whether there are any sub-object meta types are defined
            if not len(self.allowedSubObjects) != 0:
                raise DBObjectExceptions.NoAllowedSubObjectMetaTypes()
    
            # Check the db for all possible sub-objects
            for subObjectType in self.allowedSubObjects:
    
                # get the info from the DB and add the found objects to Zope
                rows = self.getAllSubObjects(subObjectType)
    
                # now create a Zope object for each database entry
                for row in rows:
                    # Inititialize the sub-object
                    subObject = General.getSubObjectClass(subObjectType)(self, dbRow=row)
                    # now add the new sub-object to the folder
                    self._setObject(str(subObject.id), subObject)
                    self[subObject.id]._afterAdd(self, None)
    
                    if hasattr(subObject, 'findExistingSubObjects'):
                        self[subObject.id].findExistingSubObjects()
                            
        if REQUEST is not None:
            return self.manage_main(self, REQUEST)

          
    def deleteDBAttributes(self):
        '''deleteAttributes(self) --> Delete the object in the DB; return None'''

        for id in self.objectIds():
            if hasattr(self[id], 'deleteDBAttributes'):
                self[id].deleteDBAttributes()


    def manage_delObjects(self, ids=[], REQUEST=None):
        '''manage_delObjects(self, ids) --> delete the objects passed in the id; return None'''

        # for every id passed for deletion do that
        for id in ids:
            # check whether the id actually exists; raise an exception if necessary
            if id in self.objectIds():
                self[id]._beforeDelete(self, REQUEST)
                # sometimes we want to delete a non-DBObject object, so do not always execute
                # 'deleteDBAttributes'
                if hasattr(self[id], 'deleteDBAttributes'):
                    self[id].deleteDBAttributes()
                # Since volatile objects do not have a ZODB presence, we do not need to delete
                # them in the ZODB
                if not self.isVolatile:
                    self._delObject(id)
                self._afterDelete(self, REQUEST)
            else:
                    raise "Delete: Object %s not found in object %s" %(id, self.id)        

        if REQUEST:
            return self.manage_main(self, REQUEST)
        

    def queryObjects(self, con, metaType=None, context=None):
        '''queryObjects(self, conditions) --> query an object folder for the correct object;
        return list of objects'''
        if not metaType:
            metaType = self.allowedSubObjects[0]

        if not context:
            context = self

        con = con + ({'attributeName': 'containertype', 'value': self.aq_parent.meta_type},
                     {'attributeName': 'containerid',   'value': self.aq_parent.id})

        dbContainer = self._getSQLMethodContainer(self)
        dbRows = dbContainer.qry_queryObjects(objectMetaType=metaType, conditions=con)

        return self.convertSQLResultToObjects(dbRows, metaType, context)


    persistent_setOb = ObjectManager._setOb
    def _setOb(self, id, object):
        '''_setOb(self, id, object) --> add the object to this folder (w/o error handling); return None'''
        if self.isVolatile or object.isVolatile:
            pass
        else:
            return self.persistent_setOb(id, object)


    persistent_setObject = ObjectManager._setObject
    def _setObject(self,id,object,roles=None,user=None, set_owner=1):
        '''_setObject(self,id,object,roles=None,user=None, set_owner=1) --> add the object to the
           container; return None'''
        if self.isVolatile or object.isVolatile:
            pass
        else:
            self.persistent_setObject(id, object, roles, user, set_owner)


    persistent_delOb = ObjectManager._delOb
    def _delOb(self, id):
        '''_setOb(self, id, object) --> delete the object from this folder (w/o error handling); return None'''
        try:
            self.persistent_delOb(id)
        except:
            pass
        

    persistent_delObject = ObjectManager._delObject
    def _delObject(self, id, dp=1):
        '''_delObject(self, id, dp=1) --> delete the object from this folder; return None'''
        try:
            return self.persistent_delObject(id, dp)
        except:
            pass


    persistent_getOb = ObjectManager._getOb
    def _getOb(self, id, default=General._marker):
        '''_getOb(self, id, default=General._marker) --> get the object; return object'''
        return self._getObject(id, default)
        

    def _getObjectSafely(self, name, default=General._marker):
        '''getObjectSafely(self, name) --> get object from the db; if not throw an informative error;
           return object or HTML'''
        try:
            return self._getObject(name, default)
        except (AttributeError, ValueError): 
            HTTPResponse().notFoundError(name)


    def _getObject(self, name, default=General._marker):
        '''_getObject(self, id, default=General._marker) --> get object from the database; return object'''        
        object = None

        if not (self.isVolatile and General.canStringBeInt(name)):
            # first try to get the object from the ZODB, in case the object is persistent
            object = self.persistent_getOb(str(name), default)
            if not object: object = None
        
        if (object is None):
            # okay, this failed, so let's see, if it exists in the DB
            dbContainer = self._getSQLMethodContainer(self)

            # since we have no way of determining the meta_type, we need to check all of them
            # Note: Ususally this is only one, so the performance will be fine!
            for metaType in self.allowedSubObjects:
                objectClass = General.getSubObjectClass(metaType)
                try: dbRows = getattr(dbContainer, 'qry_get'+objectClass.dbTable.name)(id=name)
                except: dbRows = None

                # If a result was returned, then we got something; now we need to check, whether
                # the object volatile. If so, return it!
                if dbRows and objectClass.isVolatile:
                    object = objectClass(self, dbRow=dbRows[0])
                    object._p_jar = self._p_jar
                    object = object.__of__(self)
                    object._afterAdd(self)

        if (object is None):
            if default is General._marker:
                raise AttributeError, id
            return default

        return object


    persistent_objectIds = ObjectManager.objectIds
    def objectIds(self, spec=None):
        '''objectIds(self, spec=None) --> return the ids of all objects in the container;
           return list'''
        return map(lambda ob: ob.id, self.objectValues(spec))
    

    def objectValues(self, spec=None):
        '''objectValues(self, spec=None) --> return the objects of this container; return list'''
        objects = []
        
        # first we need to prepare the metaTypes selector, since 'spec' can be None, a string or
        # a list
        if not spec:
            metaTypes = self.allowedSubObjects
        elif type(spec) != ListType and type(spec) != TupleType:
            metaTypes = (spec,)
        else:
            metaTypes = spec

        for metaType in metaTypes:
            # first get the class of the meta type, so that we can find out, whether we have a
            # volatile object class or not.
            cl = General.getSubObjectClass(metaType)
            if cl.isVolatile:
                # if the objects are volatile, then look in the database and convert the RDB
                # result values to objects
                dbRows  = self.getAllSubObjects(metaType, orderBy='id')
                for row in dbRows:
                    object = cl(self, dbRow=row)
                    object.setTemporaryAttributes(row)
                    object._p_jar = self._p_jar
                    object = object.__of__(self)
                    object._afterAdd(self)
                    objects.append(object)
            else:
                # okay, we have a persistent object, so we can use the normal methods
                objects = objects + map(self.persistent_getOb, self.persistent_objectIds(metaType))

        return objects
    
            
    def objectItems(self, spec=None):
        '''objectItems(self, spec=None) --> return a list of (id, subobject) tuples; return tuple'''
        # The original Zope API designs this method differently, but it is more efficient
        # this way, since I don't know whether an object is volatile or not before hand
        objects = self.objectValues(spec)
        return map(lambda ob: (ob.id, ob), objects)


    getId__roles__ = None
    def getId(self):
        '''getId(self) --> get id of this object; return string'''
        return self.id


    def tpURL(self):
        '''tpURL(self) --> get id of this object; return string'''
        return self.id


Globals.default__class_init__(DBObjectFolder)


