# This code is copyrighted by bluedynamics.com in June, 2001.
#
# written by Philipp Auersperg
# Last updated: June 5, 2001

import xmlrpclib


class GeneralXMLRPCSupport:
    '''GeneralXMLRPCSupport class - Defines methods that can be used by all objects'''

    __ac_permissions__ = (
    ('View DBObjects', ('getAbsoluteURL', 'returnAsDicts', 'returnXML',)), 
    )

    # Note I changed the name of this method! Why do you have this method in the first place?
    def getAbsoluteURL(self):
        '''getAbsoluteUrl(self) --> returns the absolute url; return string'''
        return self.absolute_url()


    def raiseFault(self, code, data):
        '''raises an xmlrpc Fault according to Frederik Lundh\'s xmlrpclib'''
        fc = {'faultCode': code, 'faultString': data}
        raise apply(xmlrpclib.Fault, (), fc)


    def returnAsDicts(self, methodName, args=(), kwargs={}):
        '''returnAsDicts(self, methodName, args=(), kwargs={} --> returns the object-based results of
           a method as an array of structs; return list'''
        method = getattr(self, methodName)
        objects = apply(method, args, kwargs)
    
        return map(lambda object: object.getValues(), objects)


    def returnXML(self, methodName, args=(), kwargs={}):
        '''returnXML(self, methodName, args=(), kwargs={}) --> returns the object-based result of a
           method as a struct in XML; return object'''
        method = getattr(self, methodName)
        object = apply(method, args, kwargs)
        
        if object == None:
            return xmlrpclib.Boolean(0)
        
        return object.getValues()



class DBObjectXMLRPCSupport(GeneralXMLRPCSupport):
    '''DBObjectXMLRPCSupport class - This class adds XML-RPC functionality to any DBObject'''

    __ac_permissions__ = ( (
    ('View DBObjects', ('getValue', 'getColumns', 'getValues',)),
    ('Add DBObjects', ('addObject',)),
    ('Edit DBObjects', ('update',)), )
    + GeneralXMLRPCSupport.__ac_permissions__ )
  
    def getValue(self, name):
        '''getValue(self, name) --> returns the DB Attribute with specified name; return object'''
        atts = self.getTemporaryAttributes()
        return atts[name]

  
    def getColumns(self):
        '''getColumns(self) --> returns the defined attributes for the DBObject; return list '''
        return map(lambda attr: attr.name, self.dbTable.getAllAttributes()) + ['id']

  
    def getValues(self):
        '''getValues(self) --> returns a dictionary containing the values; return dictionary'''
        result = {}
    
        for column in self.getColumns():
            result[column] = self.getTemporaryAttributes()[column]
      
        return result


    def update(self,values):
        '''update(self,values) --> performs an update of the attributes; return None'''
        self.manage_edit(values, 0)

  
    def addObject(self, metaType, kwargs):
        '''addObject(self, metaType, kwargs) --> creates a new subobject of type metaType and returns
           it\'s new ID; return int'''
        return self.manage_add(kwargs, subObjectMetaType=metaType, returnId=1)
  
      
    
class DBObjectFolderXMLRPCSupport(GeneralXMLRPCSupport):
    '''DBObjectFolderXMLRPCSupport class - This class adds XML-RPC functionality to any DBObject Folder'''
  
    __ac_permissions__ = ( (
    ('View DBObjects', ('getObjectValues',)), )
    + GeneralXMLRPCSupport.__ac_permissions__ )

    def getObjectValues(self):
        '''getObjectValues(self) --> get all the object values as dictionaries; return list
           of dictionaries'''
        result = []
        for object in self.objectValues():
            result.append(object.getValues())
      
        return result

