# This code is copyrighted by Stephan Richter in August, 2000.
#
# written by Stephan Richter
# Last updated: September 20, 2000

from Products.ZSQLMethods.SQL import SQL
from Acquisition              import Implicit
from General                  import getSubObjectClass

import string

# This value should never be inserted to the DB; it is necessary so that we can
# generate the automatic SQL Methods.
NOVALUE = ' '

class DBTable(Implicit):
    '''The DBTable class should be able to describe a DB table and generate some
       generic code, like create table, insert, update and delete rows.''' 


    def __str__(self):
        result = "Table: %s(%s)\n" %(self.name, string.join(self.inheritedTables, ', '))
        result = result + 'Attributes:\n'
        result = result + string.join(self.dbAttributes.keys(), ',\n')
        return result
    
    
    def __call__(self):
        return self.__str__()
    
    
    def __len__(self):
        return len(self.dbAttributes.keys())
    
    
    def __getitem__(self, key):
        return self.dbAttributes[key]
    
    
    def __setitem__(self, key, value):
        self.dbAttributes[key] = value
    
    
    def __delitem__ (self, key):
        del(self.dbAttributes[key])


    def __add__(self, other):
        
        newDBTable = DBTable(self.name, self.inheritedTables, self.dbAttributes.copy(),
                             self.connectionId)

        for key in other.dbAttributes.keys():
            newDBTable.dbAttributes[key] = other.dbAttributes[key]

        return newDBTable


    def __init__(self, tableName, inheritedTables=[], dbAttributes={}, connectionId='dbConnection'):
        self.name = tableName
        self.inheritedTables = inheritedTables
        self.connectionId = connectionId
        self.dbAttributes = dbAttributes


    def getAllAttributes(self):
        '''getAllAttributes(self) --> get all attributes from the local and inherited table'''
        attrs = self.dbAttributes.values()

        for table in self.inheritedTables:
            cl = getSubObjectClass(table)
            attrs = attrs + cl.dbTable.dbAttributes.values()

        return attrs


    def getAllAttributeNames(self):
        '''getAllAttributeNames(self) --> get all attributes names from the local and inherited table'''
        attrs = self.dbAttributes.keys()

        for table in self.inheritedTables:
            cl = getSubObjectClass(table)
            attrs = attrs + cl.dbTable.dbAttributes.keys()

        return attrs


    def _createIdAttribute(self, historyTable=0):

        if historyTable:
	    id = 'historyid'
	else:
	    id = 'id'
	    
	return DBAttribute(id, 'int8', zopeType='int', isPrimaryKey=1,
                           defaultValue="NEXTVAL('seq%s')" %self.name)


    def _createArgumentsString(self):

        argList = map(lambda x: x.name+'="%s"' % NOVALUE, self.getAllAttributes())
        return string.join(argList, '\n')


    def getCreateSequenceSQL(self, start=1, increment=1):
        if not self.inheritedTables:
            sql = '''CREATE SEQUENCE seq%s START %i INCREMENT %i\n''' %(self.name, start, increment)
        else:
            sql = '''CREATE VIEW seq%s AS SELECT * FROM seq%s\n''' %(self.name, self.inheritedTables[0])
        sql = sql + '<dtml-var sql_delimiter>\n'
        return sql


    def getDropSequenceSQL(self):
        if not self.inheritedTables:
            sql = '''DROP SEQUENCE seq%s\n''' %(self.name)
        else:
            sql = '''DROP VIEW seq%s\n''' %(self.name)
        sql = sql + '<dtml-var sql_delimiter>\n'
        return sql
    

    def getCreateTableSQL(self, historyTable=0):
        if not self.inheritedTables:
            attrCreateStrings = map(lambda x: x.getCreateString(), [self._createIdAttribute(historyTable)] + 
	                                                           self.dbAttributes.values())
        else:
            attrCreateStrings = map(lambda x: x.getCreateString(), self.dbAttributes.values())
        
        sql =       'CREATE TABLE %s (\n' %self.name
        sql = sql + string.join(attrCreateStrings, ',\n')
        if len(self.inheritedTables):
            sql = sql + ')\n INHERITS ( %s )\n' %string.join(self.inheritedTables, ', ')
        else:
            sql = sql + ')\n'

        sql = sql + '<dtml-var sql_delimiter>\n'
        return sql
    

    def getDropTableSQL(self):
        sql = '''DROP TABLE %s\n''' %(self.name)
        sql = sql + '<dtml-var sql_delimiter>\n'
        return sql


    def getInsertRowSQL(self):
        attrDeclarationStrings = map(lambda x: x.getInsertDeclarationString(), self.getAllAttributes())
        attrValueStrings = map(lambda x: x.getInsertValueString(), self.getAllAttributes())

        sql =       '''<dtml-call "REQUEST.set('comma', '')">\n'''
        sql = sql +   'INSERT INTO %s (\n' %self.name
        sql = sql + string.join(attrDeclarationStrings, '\n') + '\n'
        sql = sql + ''') VALUES (\n<dtml-call "REQUEST.set('comma', '')">\n'''
        sql = sql + string.join(attrValueStrings, '\n') + '\n'
        sql = sql + ''')\n'''
        sql = sql + '''<dtml-var sql_delimiter>\n'''
        sql = sql + '''SELECT last_value as id FROM seq%s\n''' %(self.name)
        sql = sql + '''<dtml-var sql_delimiter>\n'''

        return sql


    def getUpdateRowSQL(self):
        attrValueStrings = map(lambda x: x.getUpdateValueString(), self.getAllAttributes())

        sql =       '''<dtml-call "REQUEST.set('comma', '')">\n'''
        sql = sql +   'UPDATE %s SET\n' %self.name
        sql = sql + string.join(attrValueStrings, '\n') + '\n'
        sql = sql + '\nWHERE <dtml-sqltest id type="int">\n'
        sql = sql + '''<dtml-var sql_delimiter>\n'''

        return sql
        

    def getDeleteRowSQL(self):
        sql = '''DELETE FROM %s WHERE <dtml-sqltest id type="int">\n''' %(self.name)
        sql = sql + '''<dtml-var sql_delimiter>\n'''
        return sql


    def getSelectRowSQL(self):
        sql = '''SELECT * FROM %s* WHERE <dtml-sqltest id type="int">\n''' %(self.name)
        sql = sql + '''<dtml-var sql_delimiter>\n'''
        return sql


    def getGetAllSubObjectsSQL(self):
        return _getAllSubObjectsSQL


    def getGetParentInformationSQL(self):
        return _getParentInformation


    def getQueryObjectsSQL(self):
        return _queryObjects


    def makeInitTableSQLMethod(self, historyTable=0):

        sql =       '<dtml-if "dropFirst == 1">\n'
        sql = sql + '  %s\n' %self.getDropSequenceSQL()
        sql = sql + '  %s\n' %self.getDropTableSQL()
        sql = sql + '</dtml-if>\n\n'
        sql = sql + '%s\n\n' %self.getCreateSequenceSQL()
        sql = sql + '%s\n' %self.getCreateTableSQL(historyTable)

        return SQL('sql_init%s' %self.name, 'SQL: Init %s table and sequence' %self.name,
                   self.connectionId, 'dropFirst:int=1', sql)


    def makeInsertRowSQLMethod(self):
        
        return SQL('sql_add%s' %self.name, 'SQL: Insert a row into %s' %self.name,
                   self.connectionId, self._createArgumentsString(), self.getInsertRowSQL())


    def makeUpdateRowSQLMethod(self):
        
        return SQL('sql_edit%s' %self.name, 'SQL: Update a row in %s' %self.name,
                   self.connectionId, self._createArgumentsString() + '\nid=0', self.getUpdateRowSQL())


    def makeDeleteRowSQLMethod(self):
        
        return SQL('sql_delete%s' %self.name, 'SQL: Delete a row in %s' %self.name,
                   self.connectionId, 'id=0', self.getDeleteRowSQL())


    def makeSelectRowSQLMethod(self):
        
        return SQL('qry_get%s' %self.name, 'Query: Select a row in %s' %self.name,
                   self.connectionId, 'id=0', self.getSelectRowSQL())


    def makeGetAllSubObjectsSQLMethod(self):
        
        return SQL('qry_getAllSubObjects', 'Query: Get all Sub-Objects in table given container params',
                   self.connectionId, 'table containerid=0 containertype="" orderBy=""', self.getGetAllSubObjectsSQL())


    def makeGetParentInformationSQLMethod(self):
        
        return SQL('qry_getParentInformation', 'Query: Get Parent Information of an object',
                   self.connectionId, 'objectMetaType="" id=""', self.getGetParentInformationSQL())


    def makeQueryObjectsSQLMethod(self):
        
        return SQL('qry_queryObjects', 'Query: Query for the Objects',
                   self.connectionId, 'objectMetaType="" conditions', self.getQueryObjectsSQL())

    
    def createBasicSQLMethods(self, sqlMethodContainer):

        # first add the SQL Methods that come from the inherited objects (by the meaning of a SQL table - 
        # this is only a PostGreSQL feature)
        for table in self.inheritedTables:
            cl = getSubObjectClass(table)
            cl.dbTable.createBasicSQLMethods(sqlMethodContainer)

        if not hasattr(sqlMethodContainer, 'sql_init%s' %self.name):
            setattr(sqlMethodContainer, 'sql_init%s' %self.name, self.makeInitTableSQLMethod())

        if not hasattr(sqlMethodContainer, 'sql_add%s' %self.name):
            setattr(sqlMethodContainer, 'sql_add%s' %self.name, self.makeInsertRowSQLMethod())

        if not hasattr(sqlMethodContainer, 'sql_edit%s' %self.name):
            setattr(sqlMethodContainer, 'sql_edit%s' %self.name, self.makeUpdateRowSQLMethod())

        if not hasattr(sqlMethodContainer, 'sql_delete%s' %self.name):
            setattr(sqlMethodContainer, 'sql_delete%s' %self.name, self.makeDeleteRowSQLMethod())

        if not hasattr(sqlMethodContainer, 'qry_get%s' %self.name):
            setattr(sqlMethodContainer, 'qry_get%s' %self.name, self.makeSelectRowSQLMethod())

        if not hasattr(sqlMethodContainer, 'qry_getAllSubObjects'):
            setattr(sqlMethodContainer, 'qry_getAllSubObjects', self.makeGetAllSubObjectsSQLMethod())

        if not hasattr(sqlMethodContainer, 'qry_getParentInformation'):
            setattr(sqlMethodContainer, 'qry_getParentInformation', self.makeGetParentInformationSQLMethod())

        if not hasattr(sqlMethodContainer, 'qry_queryObjects'):
            setattr(sqlMethodContainer, 'qry_queryObjects', self.makeQueryObjectsSQLMethod())



class DBAttribute:
    ''' Possible attributes that are useful to think about in a DB Attribute:
         - name             (string)
         - sqlType          (string)
         - zopeType         (string)
         - isPrimaryKey     (bool) --> 0 or 1
         - defaultValue     (string)
         '''


    def __str__(self):
        '''__str__(self) --> returns the string representation of the object; return string'''
        return self.getCreateString()


    def __init__(self, name, sqlType, zopeType='string', isPrimaryKey=0, defaultValue=''):
        '''__init__(self, name, sqlType, zopeType='string', isPrimaryKey=0, defaultValue='')
           --> initializes the object; return None'''
        self.name = name
        self.sqlType = sqlType
        self.zopeType = zopeType
        self.isPrimaryKey = isPrimaryKey
        self.defaultValue = defaultValue
        

    def getCreateString(self):
        '''getCreateString(self) --> returns a string that is used to describe an attribute when
           creating a table; return string'''
        sql =       '  %s' %self.name
        sql = sql + ' '*(30-len(self.name))
        sql = sql + ' %s' %string.upper(self.sqlType)
        if self.defaultValue:
            sql = sql + ' DEFAULT %s' %self.defaultValue
        sql = sql + ' PRIMARY KEY'*self.isPrimaryKey

        return sql


    def getInsertDeclarationString(self):
        '''getInsertDeclarationString(self) --> creates the string that is used when an attribute is declared
           to be in the insert row statement; return string'''
        sql =       '''  <dtml-if "%s != '%s' and %s != _.None">\n''' %(self.name, NOVALUE, self.name)
        sql = sql + '''    <dtml-var "REQUEST['comma']">'''
        sql = sql + '''%s\n''' %self.name
        sql = sql + '''    <dtml-call "REQUEST.set('comma', ',')">\n'''
        sql = sql +   '  </dtml-if>'

        return sql


    def getInsertValueString(self):
        '''getInsertValueString(self) --> creates the imsert statement for this attribute; return string'''
        sql =       '''  <dtml-if "%s != '%s' and %s != _.None">\n''' %(self.name, NOVALUE, self.name)
        sql = sql + '''    <dtml-var "REQUEST['comma']">'''
        sql = sql + '''<dtml-sqlvar %s type="%s">\n''' %(self.name, self.zopeType)
        sql = sql + '''    <dtml-call "REQUEST.set('comma', ',')">\n'''
        sql = sql +   '  </dtml-if>'

        return sql


    def getUpdateValueString(self):
        '''getUpdateValueString(self) --> creates the update statement for this attribute; return string'''
        sql =       '''  <dtml-if "%s != '%s' and %s != _.None">\n''' %(self.name, NOVALUE, self.name)
        sql = sql + '''    <dtml-var "REQUEST['comma']">'''
        sql = sql + '''%s = <dtml-sqlvar %s type="%s">\n''' %(self.name, self.name, self.zopeType)
        sql = sql + '''    <dtml-call "REQUEST.set('comma', ',')">\n'''
        sql = sql +   '  </dtml-if>'

        return sql


_getAllSubObjectsSQL = '''
SELECT *
  FROM &dtml-table;
  WHERE ( <dtml-sqltest containerid   type="int"> AND
          <dtml-sqltest containertype type="string"> )

    <dtml-if "not (containerid and containertype)">
      OR
        ( containerid = NULL OR
          containertype = NULL )
    </dtml-if>

  <dtml-if "orderBy != ''">
    ORDER BY &dtml-orderBy;
  </dtml-if>
'''


_getParentInformation = '''
SELECT containertype, containerid
FROM <dtml-var objectMetaType>
WHERE
  <dtml-sqltest id type="int">
'''


_queryObjects = '''
SELECT *
FROM <dtml-var objectMetaType>
WHERE
  <dtml-in conditions mapping>
    <dtml-var attributeName>
    <dtml-if "_.has_key('operator')">
      <dtml-var operator>
    <dtml-else>
      =
    </dtml-if>
    '<dtml-var value>'
    <dtml-unless sequence-end>
      AND
    </dtml-unless>
  </dtml-in>
'''
