# This code is copyrighted by iuveno AG in May, 2001.
#
# written by Stephan Richter
# Last updated: May 18, 2001

# This support class will provide methods that will convert the resulting SQL field results to their
# actual objects in a given context. I also implemented some other methods that are used on the way
# to provide some other nice info: So you can for example get a list of all the parent ids in the
# correct order (using getParentPathList) or even the relative URL from the root of the DBObjects to
# the actual objects (using getRelativeObjectURL).
# I think this Support class will be extended for the upcoming SmartObjects.

import string, General


class RDB2ObjectSupport:
    '''RDB2ObjectSupport - Provides methods to convert SQL information to the actual object tree in Zope'''

    
    def getParentInformation(self):
        '''getParentInformation(self) --> get the parent\'s object information; return tuple'''

        # get the id and the meta type of the parent object; This method is good, if the object is not
        # inside the aquisition tree anymore, and we still have to figure out its parent object's info
        container = self._getSQLMethodContainer(self)
        dbRows = container.qry_getParentInformation(objectMetaType=self.meta_type, id=self.id)

        return dbRows[0].containertype, str(dbRows[0].containerid)


    def getParentPathList(self, idList=None, containertype=None, containerid=None):
        '''getParentPathList(self) --> get a list of all parents in the correct order; return list '''

        # make an idList, if none was passed. The idList contains the ids of the objects that are
        # in the path of the concerned object 
        if idList == None:
            idList = [self.id, self.meta_type+'s']

        # get the container type and id, if none were passed
        if not (containertype and containerid):
            containertype, containerid = self.getParentInformation()

        sqlContainer = self._getSQLMethodContainer(self)

        # Okay, this is recurrsion without a recursive method. Pfui!!! (But SQL can't do it!)
        while containertype not in ('', None) and containerid not in ('', None, 0):

            # So, no add the DBObject id and its Folder to the list 
            idList.append(containerid)
            idList.append(containertype+'s')

            # now fetch the new container info
            dbRows = sqlContainer.qry_getParentInformation(objectMetaType=containertype, id=containerid)
            containertype, containerid = dbRows[0].containertype, dbRows[0].containerid

        # reverse the list, so it has the right order ;)
        idList.reverse()

        # Since usually the root object is going to have a non-standard id, we are going to replace it with the real
        # one, since this one is easy to grep
        idList[0] = sqlContainer.id

        return idList


    def getRelativeObjectURL(self, idList=None, containertype=None, containerid=None):
        '''getRelativeObjectURL(self) --> generate the relative (up to the DBObject Root) URL of the tree;
           return string'''
        # join the ids together to a partial URL
        return string.join(self.getParentPathList(idList, containertype, containerid), '/')+'/'


    def getObjectTreeWithSQL(self, context, idList=None, containertype=None, containerid=None):
        '''getObjectTreeWithSQL(self) --> get the true object tree, by using the references in the SQL tables;
           return object'''
        object = context

        # So, get all the ids in a lisy
        if containertype and containerid:
            list   = self.getParentPathList(idList, containertype, containerid)
        else:
            list = idList
            list.reverse()

        # Now we go through every id and try to get the actual object
        if list[0] == object.id:
            list.remove(list[0])

        for id in list:
            object = object[id]

        return object


    def convertSQLResultToObjects(self, dbRows, metaType=None, context=None):
        '''convertSQLResultToObjects(self, dbRows, metaType=None, context=None) --> convert the SQL Results
           to a nice list of objects; return list'''

        # Setup iniital object context, if none was passed
        if not context:
            context = self

        # set a meta type of the resulting objects, if it was not passed
        if not metaType:
            metaType = self.allowedSubObjects[0]

        result = []

        # for each row find the object
        for row in dbRows:
            object = self.getObjectTreeWithSQL(context, [row.id, metaType+'s', ],
                                               row.containertype, row.containerid)
            object.setTemporaryAttributes(row)
            result.append(object)

        return result


    # *** the good way of doing it ;-) ***

    def makeObjectFromRow(self, row, parent, metaType=None):
        '''makeObjectFromRow(self, row, parent, metaType=None) --> make an object from a DB row;
           return object'''
        if not metaType: cl = self.__class__
        else: cl = General.getSubObjectClass(metaType)

        object = cl(parent, dbRow=row)
        cl._p_jar = parent._p_jar
        object = object.__of__(parent)
        object._afterAdd(parent)
        return object
            

    def makeObjectsFromRows(self, rows, parent , metaType=None):
        '''makeObjectsFromRows(self, rows, parent, metaType=None) --> make an object list from
           a DB rows; return object'''
        if not metaType: cl = self.__class__
        else: cl = General.getSubObjectClass(metaType)

        objects = []
        for row in rows:
            object = cl(parent, dbRow=row)
            cl._p_jar = parent._p_jar
            object = object.__of__(parent)
            object._afterAdd(parent)
            objects.append(object)

        return objects

