# This code is copyrighted by iuveno AG in January, 2001.
#
# written by Stephan Richter
# Last updated: February 4, 2001

try:
    from crypt import crypt
except ImportError:
    crypt = None
import sha, binascii, string

from Globals            import HTMLFile

from Products.DBObjects.DBTable import DBTable, DBAttribute
from AccessControl.User         import BasicUser
from AccessControl              import getSecurityManager


class UserSupport(BasicUser):
    '''UserSupport class - implements a User Folder for DB Objects'''

    dbTable = DBTable('UserSupport', [],
                      {'username'      : DBAttribute('username',       'varchar(50)' ),
                       'password'      : DBAttribute('password',       'varchar(50)' ),
                       'roles'         : DBAttribute('roles',          'text'        )})

    manage_viewRoles = HTMLFile('user_viewRoles', globals())

    __ac_permissions__ = (('View DBObjects',  ('manage_viewRoles',)),) 

    manage_options = ({'label':'Roles',        'action':'manage_viewRoles'},)

    # this attribute defines the variables that will be responsible for the authentication
    # mechanism; since some databases (such as Interbase) have 'username', 'password' und
    # 'roles' as keywords, you can simply change these dictionary entries to have different
    # variable names
    authAttributes = {'username': 'username', 'password': 'password', 'roles': 'roles'}


    ###  Zope User-specific API methods  ###

    def getUserName(self):
        '''getUserName(self) --> get the username of the member; return string'''
        return self.getAttributes()[0][self.authAttributes['username']]


    def _getPassword(self):
        '''_getPassword(self) --> get the password of the member; return string'''
        return self.getAttributes()[0][self.authAttributes['password']]   


    def getRoles(self):
        '''getRoles(self) --> get the roles of the member; return list'''
        return self._getAllRoles()


    def getDomains(self):
        '''getDomains(self) --> get the domains of the member; return list'''
        return []


    def authenticate(self, password, request):
	'''authenticate(self, password, request) --> authenticate user; return boolean'''
	passwrd = self._getPassword()
	result  = 0
	if (passwrd[2:5]) == '{sha}':
	    password = binascii.b2a_base64(sha.new(password).digest())[:-1]
	    result = passwrd[5:] == password
	elif (passwrd[:7]) == '{crypt}' and crypt is not None:
	    #if crypt is None, it's not compiled in and everything will fail
	    password = crypt(password, passwrd[7:9])
	    result = passwrd[7:] == password
	else:
	    result = passwrd == password
        
	domains=self.getDomains()
        if domains:
            return result and domainSpecMatch(domains, request)
        return result

    ###  End of API methods  ###


    def addRoles(self, newRoles):
        '''Add roles to the user'''
        roles = filter(None, string.split(self.getAttributes()[0][self.authAttributes['roles']], ','))
        roles = roles + newRoles
        dict = {'roles': string.join(roles, ',')}
        self.manage_edit(dict, 0)
        

    def _getGroupRoles(self):
        '''getGroupRoles(self) --> Get roles that are due to group membership; return list of strings'''
        return self.getAllInheritedGroupRoles()


    def _getPersonalRoles(self):
        '''_getPersonalRoles(self) --> get personal roles; return list of strings'''
        roles = ['Person', 'Person%s' %self.id]
        return roles


    def getSpecialRoles(self):
        '''getSpecialRoles(self) --> get special Roles of Person; return list '''
        return []


    def _getAllRoles(self):
        '''_getAllRoles(self) --> Get all the roles of a person; return list of strings'''
        roles = filter(None, string.split(self.getAttributes()[0][self.authAttributes['roles']], ','))
        roles = roles + self.getSpecialRoles()
        roles = roles + self._getPersonalRoles()
        roles = roles + self._getGroupRoles()
        return roles


    def canAccessObject(self, object, type='view'):
        '''canAccessObject(self, object, type=\'view\') --> check whether the user can access a certain object;
           return boolean'''

        securityManager = getSecurityManager()
        
        if type == 'view':
            try:
                return securityManager.validate(object, object, 'hasViewPermission', object.hasViewPermission)
            except:
                return 0

        elif type == 'manage':
            try:
                return securityManager.validate(object, object, 'hasManagePermission', object.hasViewPermission)
            except:
                return 0

        else:
            return 0
        









