# This code is copyrighted by Stephan Richter in September, 2001.
#
# written by Stephan Richter
# Last updated: September 2, 2001

from OFS.Folder         import Folder
from OFS.DTMLMethod     import addDTMLMethod
from Products.DBObjects import DBObjectRoleSupport

class ZopeFolder(Folder, DBObjectRoleSupport):
    '''A DBObject API compliant ZODB Folder.'''

    meta_type = 'ZopeFolder'  
    icon = 'misc_/DBObjects/zopeFolder'

    # API-compliant variables
    isVolatile = 0
    subObjectIds = ()
    allowedSubObjects = ()
    
    manageRoles = ('ManageZopeFolders', 'ManageZopeFolder%(zopefolderid)s',)
    viewRoles   = ('ViewZopeFolders', 'ViewZopeFolder%(zopefolderid)s')

    __ac_permissions__=(
        ('View management screens', ('manage_main',) ),
        ('Add DBObjects',           ('hasManagePermission',) ),
        ('Edit DBObjects',          ('hasManagePermission',) ),
        ('Delete DBObjects',        ('deleteDBAttributes', 'hasManagePermission')),
        ('View DBObjects',          ('getAttributes', 'hasViewPermission') ),
        )

    def __init__(self, parent, id, title):
        ''' API compliant __init__ method'''
        self.id = id
        addDTMLMethod(self, 'index_html', 'Main Page', _index_html)
        

    def deleteDBAttributes(self):
        '''DBObject API compliance'''
        pass


    def findExistingSubObjects(self, REQUEST=None):
        '''DBObject API compliance '''
        pass


    def hasViewPermission(self):
        '''hasViewPermission(self) --> execute method to see whether user has view rights; return boolean'''
        return 1


    def hasManagePermission(self):
        '''hasManagePermission(self) --> execute method to see whether user has manage rights; return boolean'''
        return 1


_index_html = '''
<dtml-var standard_html_header>
  <h1>This is the Default View of this ZopeFolder!</h1>
<dtml-var standard_html_footer>
'''
